import lvlspy.props as lp
import numpy as np
from astropy import units as u

units_dict = {"eV": 1000, "keV": 1, "MeV": 1.0e-3, "GeV": 1.0e-6}


class Level(lp.Properties):
    """A class for storing and retrieving data about a level.

    Args:
        ``energy`` (:obj:`float`): The energy of the level.  It is in units
        given by the keyword `units`.

        ``multiplicity`` (:obj:`int`): The multiplicity of the level.

        ``units`` (:obj:`str`, optional):  A string giving the
        units for the energy.

    """

    def __init__(self, energy, multiplicity, units="keV"):
        self.energy = energy / units_dict[units]
        self.multiplicity = multiplicity
        self.properties = {}
        self.units = "keV"

    def __eq__(self, other):
        if not isinstance(other, Level):
            return NotImplemented

        return self.energy == other.energy and self.multiplicity == other.multiplicity

    def get_energy(self, units="keV"):
        """Method to retrieve the energy for a level.

        Args:
            ``units`` (:obj:`str`, optional):  A string giving the
            units for the energy.

        Returns:
            :obj:`float`: The energy.

        """

        return units_dict[units] * self.energy

    def get_multiplicity(self):
        """Method to retrieve the multiplicity for a level.

        Returns:
            :obj:`int`: The multiplicity.

        """

        return self.multiplicity

    def update_energy(self, energy, units="keV"):
        """Method to update the energy for a level.

        Args:
            ``energy`` (:obj:`float`):  The new energy for the level.

            ``units`` (:obj:`str`, optional):  A string giving the
            units for the energy.

        Returns:
            On successful return, the energy has been updated.

        """

        self.energy = units_dict[units] * energy

    def update_multiplicity(self, multiplicity):
        """Method to update the multiplicity for a level.

        Args:
            ``multiplicity`` (:obj:`int`):  The new multiplicity for the level.

        Returns:
            On successful return, the multiplicity has been updated.

        """

        self.multiplicity = multiplicity

    def compute_Boltzmann_factor(self, T):
        """Method to compute the Boltzmann factor for a level.

        Args:
            ``T`` (:obj:`float`):  The temperature in K at which to compute
            the factor.

        Returns:
            :obj:`float`: The computed Boltzmann factor
            multiplicity * exp(-Energy/kT).

        """

        T_K = T * u.K
        T_keV = T_K.to(u.keV, equivalencies=u.temperature_energy())

        return self.multiplicity * np.exp(-self.energy / T_keV.value)



import numpy as np
import lvlspy.props as lp
from astropy import units as u
from astropy.constants import astropyconst20 as const


class Transition(lp.Properties):
    """A class for storing and retrieving data about a transition.

    Args:
        ``upper_level`` (:obj:`lvlspy.level.Level`) The level from which
        there is a spontaneous decay.

        ``lower_level`` (:obj:`lvlspy.level.Level`) The level to which
        there is a spontaneous decay.

        ``Einstein_A`` (:obj:`float`): The Einstein A coefficient (the spontaneous
        decay rate per second from `upper_level` to `lower_level`).

    """

    def __init__(self, upper_level, lower_level, Einstein_A):
        self.properties = {}
        self.upper_level = upper_level
        self.lower_level = lower_level
        self.Einstein_A = Einstein_A

    def __eq__(self, other):
        if not isinstance(other, Transition):
            return NotImplemented

        return (
            self.upper_level == other.upper_level
            and self.lower_level == other.lower_level
            and self.Einstein_A == other.Einstein_A
        )

    def get_upper_level(self):
        """Method to retrieve the `upper_level` for the transition.

        Returns:
            :obj:`lvlspy.level.Level`: The `upper_level` for the transition.

        """

        return self.upper_level

    def get_lower_level(self):
        """Method to retrieve the `lower_level` for the transition.

        Returns:
            :obj:`lvlspy.level.Level`: The `lower_level` for the transition.

        """

        return self.lower_level

    def get_Einstein_A(self):
        """Method to retrieve the Einstein A coefficient for the transition.

        Returns:
            :obj:`float`: The spontaneous rate (per second) for the transition.

        """

        return self.Einstein_A

    def get_Einstein_B_upper_to_lower(self):
        """Method to get the Einstein B coefficient for the upper level
        to lower level transition (induced emission).

        Returns:
            :obj:`float`: The Einstein coefficient in cm\ :sup:`2`
            steradian per erg per s.

        """

        nu = self.get_frequency()

        result = self.Einstein_A / self._fnu()

        return result

    def get_Einstein_B_lower_to_upper(self):
        """Method to get the Einstein B coefficient for the lower level
        to upper level transition (induced absorption).

        Returns:
            :obj:`float`: The Einstein coefficient in cm\ :sup:`2`
            steradian per erg per s.

        """

        return self.get_Einstein_B_upper_to_lower() * (
            self.upper_level.get_multiplicity() / self.lower_level.get_multiplicity()
        )

    def compute_lower_to_upper_rate(self, T):
        """Method to compute the total rate for transition from the lower level to
        upper level.

        Args:
            ``T`` (:obj:`float`:) The temperature in K at which to compute
            the rate.

        Returns:
            :obj:`float`: The rate (per second).

        """

        return self.get_Einstein_B_lower_to_upper() * self._bb(T)

    def compute_upper_to_lower_rate(self, T):
        """Method to compute the total rate for transition from the upper level to
        to lower level.

        Args:
            ``T`` (:obj:`float`:) The temperature in K at which to compute
            the rate.

        Returns:
            :obj:`float`: The rate (per second).

        """

        return self.get_Einstein_A() + self.get_Einstein_B_upper_to_lower() * self._bb(
            T
        )

    def get_frequency(self):
        """Method to compute the frequency of the transition.

        Returns:
            :obj:`float`: The frequency (in Hz) of the transition.

        """

        deltaE = self.upper_level.get_energy() - self.lower_level.get_energy()

        deltaE_erg = (deltaE * u.keV).to("erg")

        return (deltaE_erg / const.h.cgs).value

    def _fnu(self):
        return (
            2.0
            * const.h.cgs
            * np.power(self.get_frequency() * u.Hz, 3)
            / np.power(const.c.cgs, 2)
        ).value

    def _bb(self, T):
        T_K = T * u.K
        T_keV = T_K.to(u.keV, equivalencies=u.temperature_energy())

        deltaE = self.upper_level.get_energy() - self.lower_level.get_energy()

        x = (deltaE * u.keV / T_keV).value

        if x < 500:
            return self._fnu() / np.expm1(x)
        else:
            return self._fnu() * np.exp(-x)




import numpy as np
from lxml import etree
import lvlspy.props as lp
import lvlspy.species as ls
import lvlspy.level as lv
import lvlspy.transition as lt


class SpColl(lp.Properties):
    """A class for storing and retrieving data about a species collection.

    Args:
        ``species`` (:obj:`list`, optional): A list of individual
        :obj:`lvlspy.species.Species` objects.

    """

    def __init__(self, species=None):
        self.properties = {}
        self.spcoll = {}
        if species:
            for sp in species:
                self.spcoll[sp.get_name()] = sp

    def add_species(self, species):
        """Method to add a species to a collection.

        Args:
            ``species`` (:obj:`lvlspy.species.Species`) The species to be added.

        Return:
            On successful return, the species has been added.

        """

        self.spcoll[species.get_name()] = species

    def remove_species(self, species):
        """Method to remove a species from a species collection.

        Args:
            ``species`` (:obj:`lvlspy.species.Species`) The species to be removed.

        Return:
            On successful return, the species has been removed.

        """

        self.spcoll.pop(species.get_name())

    def get(self):
        """Method to retrieve the species collection as a dictionary.

        Returns:
            :obj:`dict`: A dictionary of the species.

        """

        return self.spcoll

    def write_to_xml(self, file, pretty_print=True, units="keV"):
        """Method to write the collection to XML.

        Args:
            ``file`` (:obj:`str`) The output file name.

            ``pretty_print`` (:obj:`bool`, optional): If set to True,
            routine outputs the xml in nice indented format.

            ``units`` (:obj:`str`, optional): A string for the energy units.

        Return:
            On successful return, the species collection data have been
            written to the XML output file.

        """

        root = etree.Element("species_collection")
        xml = etree.ElementTree(root)

        self._add_optional_properties(root, self)

        my_coll = self.get()

        for sp in my_coll:

            tu_dict = {}

            for transition in my_coll[sp].get_transitions():
                e_upper = (transition.get_upper_level()).get_energy()

                if e_upper not in tu_dict:
                    tu_dict[e_upper] = []

                tu_dict[e_upper].append(transition)

            my_species = etree.SubElement(root, "species", name=sp)

            self._add_optional_properties(my_species, my_coll[sp])

            my_levels = etree.SubElement(my_species, "levels")

            levels = my_coll[sp].get_levels()

            for level in levels:
                my_level = etree.SubElement(my_levels, "level")
                self._add_optional_properties(my_level, level)
                my_level_props = etree.SubElement(my_level, "properties")
                if units != "keV":
                    my_energy = etree.SubElement(my_level_props, "energy", units=units)
                else:
                    my_energy = etree.SubElement(my_level_props, "energy")
                my_energy.text = self._get_energy_text(level.get_energy(), units)
                my_multiplicity = etree.SubElement(my_level_props, "multiplicity")
                my_multiplicity.text = str(level.get_multiplicity())

                if level.get_energy() in tu_dict:

                    if len(tu_dict[level.get_energy()]) > 0:
                        my_transitions = etree.SubElement(my_level, "transitions")
                        for transition in tu_dict[level.get_energy()]:
                            my_trans = etree.SubElement(my_transitions, "transition")
                            self._add_optional_properties(my_trans, transition)
                            lower_level = transition.get_lower_level()
                            if units != "keV":
                                my_to_energy = etree.SubElement(
                                    my_trans, "to_energy", units=units
                                )
                            else:
                                my_to_energy = etree.SubElement(my_trans, "to_energy")
                            my_to_energy.text = self._get_energy_text(
                                lower_level.get_energy(), units
                            )
                            my_to_multiplicity = etree.SubElement(
                                my_trans, "to_multiplicity"
                            )
                            my_to_multiplicity.text = str(
                                lower_level.get_multiplicity()
                            )
                            my_a = etree.SubElement(my_trans, "a")
                            my_a.text = str(transition.get_Einstein_A())

        xml.write(file, pretty_print=pretty_print)

    def _get_energy_text(self, energy, units):
        return str(energy * lv.units_dict[units])

    def _add_optional_properties(self, my_element, my_object):
        my_props = my_object.get_properties()

        if len(my_props):
            props = etree.SubElement(my_element, "optional_properties")
            for prop in my_props:
                if isinstance(prop, str):
                    my_prop = etree.SubElement(props, "property", name=prop)
                elif isinstance(prop, tuple):
                    if len(prop) == 2:
                        my_prop = etree.SubElement(
                            props, "property", name=prop[0], tag1=prop[1]
                        )
                    elif len(prop) == 3:
                        my_prop = etree.SubElement(
                            props, "property", name=prop[0], tag1=prop[1], tag2=prop[2]
                        )
                else:
                    print("Improper property key")
                    exit()

                my_prop.text = my_props[prop]

    def _update_optional_properties(self, my_element, my_object):
        opt_props = my_element.xpath("optional_properties")

        if len(opt_props) > 0:
            props = opt_props[0].xpath("property")

            my_props = {}
            for prop in props:
                attributes = prop.attrib
                my_keys = attributes.keys()
                if len(my_keys) == 1:
                    my_props[attributes[my_keys[0]]] = prop.text
                elif len(my_keys) == 2:
                    my_props[
                        (attributes[my_keys[0]], attributes[my_keys[1]])
                    ] = prop.text
                elif len(my_keys) == 3:
                    my_props[
                        (
                            attributes[my_keys[0]],
                            attributes[my_keys[1]],
                            attributes[my_keys[2]],
                        )
                    ] = prop.text
                else:
                    print("Improper keys for property")
                    exit()

            my_object.update_properties(my_props)

    def validate(self, file):
        """Method to validate a species collection XML file.

        Args:
            ``file`` (:obj:`str`) The name of the XML file to validate.

        Returns:
            An error message if invalid and nothing if valid.

        """

        parser = etree.XMLParser(remove_blank_text=True)
        xml = etree.parse(file, parser)
        xml.xinclude()

        url = "http://liblvls.sourceforge.net/xsd_pub/2022-09-10/spcoll.xsd"

        xml_validator = etree.XMLSchema(file=url)
        xml_validator.assert_(xml)

    def update_from_xml(self, file):
        """Method to update a species collection from an XML file.

        Args:
            ``file`` (:obj:`str`) The name of the XML file from which to update.

        Returns:
            On successful return, the species collection has been updated.

        """

        parser = etree.XMLParser(remove_blank_text=True)
        xml = etree.parse(file, parser)
        xml.xinclude()

        spcoll = xml.getroot()

        self._update_optional_properties(spcoll, self)

        el_species = spcoll.xpath("//species")

        species = []

        for sp in el_species:
            level_dict = {}
            my_species = ls.Species(sp.attrib["name"])
            self._update_optional_properties(sp, my_species)
            el_level = sp.xpath(".//level")
            for lev in el_level:
                props = lev.xpath(".//properties")
                energy = props[0].xpath(".//energy")
                multiplicity = props[0].xpath(".//multiplicity")
                attributes = energy[0].attrib
                if "units" in attributes:
                    my_level = lv.Level(
                        float(energy[0].text),
                        int(multiplicity[0].text),
                        units=attributes["units"],
                    )
                else:
                    my_level = lv.Level(
                        float(energy[0].text), int(multiplicity[0].text)
                    )
                self._update_optional_properties(lev, my_level)

                level_dict[my_level.get_energy()] = my_level

                my_species.add_level(my_level)

                el_trans = lev.xpath(".//transition")
                for trans in el_trans:
                    to_energy = trans.xpath(".//to_energy")
                    to_multiplicity = trans.xpath(".//to_multiplicity")
                    to_a = trans.xpath(".//a")

                    f_to_energy = self._convert_to_keV(to_energy)
                    if f_to_energy in level_dict:
                        my_trans = lt.Transition(
                            my_level, level_dict[f_to_energy], float(to_a[0].text)
                        )
                        self._update_optional_properties(trans, my_trans)
                        my_species.add_transition(my_trans)

            self.add_species(my_species)

    def _convert_to_keV(self, energy):
        attributes = energy[0].attrib
        result = float(energy[0].text)
        if "units" in attributes:
            result /= lv.units_dict[attributes["units"]]
        return result



class Properties:
    """A class for storing and retrieving optional properties."""

    def __init__(self):
        self.properties = {}

    def get_properties(self):
        """Method to retrieve the optional properties.

        Returns:
            :obj:`dict`: The dictionary of current properties.

        """

        return self.properties

    def update_properties(self, properties):
        """Method to update the optional properties.

        Args:
            ``properties`` (:obj:`dict`):  A dictionary of properties.
            New properties are added.  Old properties are updated.

        Returns:
            On successful return, the properties have been updated.

        """

        self.properties = {**self.properties, **properties}



import numpy as np
import lvlspy.props as lp
import lvlspy.level as lv
import lvlspy.transition as lt


class Species(lp.Properties):
    """A class for storing and retrieving data about a species.

    Args:
        ``name`` (:obj:`str`): The name of the species.

        ``levels`` (:obj:`list`, optional): A list of individual
        :obj:`lvlspy.level.Level` objects.

        ``transitions`` (:obj:`list`, optional): A list of individual
        :obj:`lvlspy.transition.Transition` objects.

        ``units`` (:obj:`str`, optional):  A string giving the
        units for the energy.

    """

    def __init__(self, name, levels=None, transitions=None):
        self.name = name
        self.levels = []
        self.transitions = []
        self.properties = {}
        if levels:
            for level in levels:
                self.levels.append(level)
        if transitions:
            for transition in transitions:
                self.transitions.append(transition)

    def get_name(self):
        """Retrieve the name of the species.

        Return:
            The :obj:`str` giving the species name.

        """

        return self.name

    def add_level(self, level):
        """Method to add a level to a species.

        Args:
            ``level`` (:obj:`lvlspy.level.Level`) The level to be added.

        Return:
            On successful return, the level has been added.

        """

        self.levels.append(level)

    def remove_level(self, level):
        """Method to remove a level from a species.

        Args:
            ``level`` (:obj:`lvlspy.level.Level`) The level to be removed.

        Return:
            On successful return, the level has been removed.

        """

        self.levels.remove(level)

    def add_transition(self, transition):
        """Method to add a transition to a species.

        Args:
            ``transition`` (:obj:`lvlspy.transition.Transition`) The transition
            to be added.

        Return:
            On successful return, the transition has been added.

        """

        self.transitions.append(transition)

    def remove_transition(self, transition):
        """Method to remove a transition from a species.

        Args:
            ``transition`` (:obj:`lvlspy.transition.Transition`) The transition
            to be removed.

        Return:
            On successful return, the transition has been removed.

        """

        self.transitions.remove(transition)

    def get_levels(self):
        """Method to retrieve the levels for a species.

        Returns:
            :obj:`list`: A list of the levels.  The levels are sorted in
            ascending energy.

        """

        return sorted(self.levels, key=lambda x: x.energy)

    def get_transitions(self):
        """Method to retrieve the transitions for a species.

        Returns:
            :obj:`list`: A list of the transitions.

        """

        return self.transitions

    def compute_equilibrium_probabilities(self, T):
        """Method to compute the equilibrium probabilities for levels in a species.

        Args:
            ``T`` (:obj:`float`): The temperature in K at which to compute the
            equilibrium probabilities.

            Returns:
                :obj:`numpy.array`: A numpy array of the probabilities of the levels.
                The levels are sorted in ascending energy.

        """

        levs = self.get_levels()

        p = np.empty(len(levs))

        for i in range(len(levs)):
            p[i] = levs[i].compute_Boltzmann_factor(T)

        p /= np.sum(p)

        return p

    def compute_rate_matrix(self, T):
        """Method to compute the rate matrix for a species.

        Args:
            ``T`` (:obj:`float`): The temperature in K at which to compute the
            rate matrix.

            Returns:
                :obj:`numpy.array`: A 2d numpy array giving the rate matrix.

        """

        levels = self.get_levels()

        rate_matrix = np.zeros((len(levels), len(levels)))

        transitions = self.get_transitions()

        for transition in transitions:
            i_upper = levels.index(transition.get_upper_level())
            i_lower = levels.index(transition.get_lower_level())

            r_upper_to_lower = transition.compute_upper_to_lower_rate(T)
            r_lower_to_upper = transition.compute_lower_to_upper_rate(T)

            rate_matrix[i_lower, i_upper] += r_upper_to_lower
            rate_matrix[i_upper, i_upper] -= r_upper_to_lower

            rate_matrix[i_upper, i_lower] += r_lower_to_upper
            rate_matrix[i_lower, i_lower] -= r_lower_to_upper

        return rate_matrix
